from enum import Enum, auto


class AccessType(Enum):
    """Enumeration of access types. Specifies how data will be read
    and returned.

    JSON_STR: Result will be returned as an unparsed json string.
    FILE_PATH: Result will be returned as the file path to the file
    containing the data.
    OBJ: The json will be parsed and returned as a python object.
    URI: A string which is a unique identifier of this asset between
    implementations of Aerovaldb. Can be used with `get_by_uuid()` and
    `put_by_uuid()` to read or write respectively.
    (_ROW_ID: For Internal use)
    MTIME: The timestamp for last modification for the resource will be
    returned (as datetime.datetime).
    CTIME: The creation timestamp for the resource will be returned (as
    datetime.datetime)
    """

    JSON_STR = auto()
    FILE_PATH = auto()
    OBJ = auto()
    URI = auto()
    BLOB = auto()
    MTIME = auto()
    CTIME = auto()



ROUTE_GLOB_STATS = "/v0/glob_stats/{project}/{experiment}/{frequency}"

ROUTE_REG_STATS = "/v0/regional_stats/{project}/{experiment}/{frequency}"

ROUTE_HEATMAP = "/v0/heatmap/{project}/{experiment}/{frequency}"

ROUTE_CONTOUR = "/v0/contour/{project}/{experiment}/{obsvar}/{model}"

ROUTE_TIMESERIES = "/v0/ts/{project}/{experiment}/{location}/{network}/{obsvar}/{layer}"

ROUTE_TIMESERIES_WEEKLY = (
    "/v0/ts_weekly/{project}/{experiment}/{location}_{network}-{obsvar}_{layer}"
)

ROUTE_EXPERIMENTS = "/v0/experiments/{project}"

ROUTE_CONFIG = "/v0/config/{project}/{experiment}"

ROUTE_MENU = "/v0/menu/{project}/{experiment}"

ROUTE_STATISTICS = "/v0/statistics/{project}/{experiment}"

ROUTE_RANGES = "/v0/ranges/{project}/{experiment}"

ROUTE_REGIONS = "/v0/regions/{project}/{experiment}"

ROUTE_MODELS_STYLE = "/v0/model_style/{project}"

ROUTE_MAP = "/v0/map/{project}/{experiment}/{network}/{obsvar}/{layer}/{model}/{modvar}"

ROUTE_SCATTER = (
    "/v0/scat/{project}/{experiment}/{network}/{obsvar}/{layer}/{model}/{modvar}"
)

ROUTE_PROFILES = "/v0/profiles/{project}/{experiment}/{location}/{network}/{obsvar}"

ROUTE_HEATMAP_TIMESERIES = "/v0/hm_ts/{project}/{experiment}"

ROUTE_FORECAST = (
    "/v0/forecast/{project}/{experiment}/{region}/{network}/{obsvar}/{layer}"
)

ROUTE_GRIDDED_MAP = "/v0/gridded_map/{project}/{experiment}/{obsvar}/{model}"

ROUTE_REPORT = "/v0/report/{project}/{experiment}/{title}"

ROUTE_REPORT_IMAGE = "/v0/report-image/{project}/{experiment}/{path}"

ROUTE_MAP_OVERLAY = "/v0/map-overlay/{project}/{experiment}/{source}/{variable}/{date}"

ALL_ROUTES = [
    ROUTE_GLOB_STATS,
    ROUTE_REG_STATS,
    ROUTE_HEATMAP,
    ROUTE_CONTOUR,
    ROUTE_TIMESERIES,
    ROUTE_TIMESERIES_WEEKLY,
    ROUTE_EXPERIMENTS,
    ROUTE_CONFIG,
    ROUTE_MENU,
    ROUTE_STATISTICS,
    ROUTE_RANGES,
    ROUTE_REGIONS,
    ROUTE_MODELS_STYLE,
    ROUTE_MAP,
    ROUTE_SCATTER,
    ROUTE_PROFILES,
    ROUTE_HEATMAP_TIMESERIES,
    ROUTE_FORECAST,
    ROUTE_GRIDDED_MAP,
    ROUTE_REPORT,
    ROUTE_REPORT_IMAGE,
    ROUTE_MAP_OVERLAY,
]



from .asyncio import async_and_sync, has_async_loop, run_until_finished
from .json import json_dumps_wrapper
from .string_utils import str_to_bool, validate_filename_component
from .uri import build_uri, extract_substitutions, parse_formatted_string, parse_uri



import logging
import os
from collections import defaultdict, deque
from pathlib import Path
from typing import Hashable, TypedDict

import aiofile

from ..utils import async_and_sync

logger = logging.getLogger(__name__)


class LRUQueue:
    """
    Small helper class that efficiently maintains a LRUQueue
    by combining a set and deque to maintain a unique constraint
    on the queue. Re-adding an element will return it to the
    end of the queue.
    """

    def __init__(self):
        self._set = set()
        self._deque = deque()

    @property
    def size(self) -> int:
        """Returns the lenth of the queue."""
        return len(self._set)

    def add(self, item: Hashable):
        """
        Adds an item to the queue.
        """
        if item in self._set:
            self._deque.remove(item)

        self._set.add(item)
        self._deque.appendleft(item)

    def pop(self) -> Hashable:
        """Removes and returns the top item from the queue."""
        item = self._deque.pop()
        self._set.remove(item)
        return item

    def remove(self, item: Hashable):
        """
        Removes an item from the queue.
        """
        if item in self._set:
            self._set.remove(item)
            self._deque.remove(item)


class CacheEntry(TypedDict):
    json: str

    last_modified: float


class JSONLRUCache:
    """
    Implements an in-memory LRU cache for file content in aerovaldb.
    """

    def __init__(self, *, max_size: int, asyncio: bool = False):
        """
        :param max_size : The maximum size of the cache in terms of number of entries / files.

        Files will be ejected based on least recently used, when full.
        """
        self._asyncio = asyncio
        self._max_size = max_size
        self.invalidate_all()

    def invalidate_all(self) -> None:
        logger.debug("JSON Cache invalidated.")

        # Stores the actual cached content, indexed by canonical file path.
        self._cache: defaultdict[str, CacheEntry | None] = defaultdict(lambda: None)

        # Stores queue of cache accesses, used for implementing LRU logic.
        self._queue = LRUQueue()

        # Tally of cache hits and misses.
        self._hit_count: int = 0
        self._miss_count: int = 0

    @property
    def hit_count(self) -> int:
        """Returns the number of cache hits since the last `invalidate_all()` call.

        Note:
        -----
        This does not include calls with `no_cache=True`
        """
        return self._hit_count

    @property
    def size(self) -> int:
        """Returns the current size of the cache in terms of number of elements."""
        return self._queue.size

    @property
    def miss_count(self) -> int:
        """Returns the number of cache misses since the last `invalidate_all()` call.

        Note:
        -----
        This does not include calls with `no_cache=True`
        """
        return self._miss_count

    def _canonical_file_path(self, file_path: str | Path) -> str:
        """
        Returns an absolute file path with symlinks removed for a file to
        ensure correct lookup.

        :param file_path : The file path.

        :return : The file path converted to canonical file path.
        """
        return str(os.path.realpath(file_path))

    async def _read_json(self, file_path: str | Path) -> str:
        abspath = self._canonical_file_path(file_path)
        logger.debug(f"Reading file {abspath}")
        if self._asyncio:
            async with aiofile.async_open(abspath, "r") as f:
                return await f.read()

        with open(abspath, "r") as f:
            return f.read()

    def _get(self, abspath: str) -> str:
        """Returns an element from the cache."""
        self._queue.add(abspath)
        self._hit_count = self._hit_count + 1
        logger.debug(f"Returning contents from file {abspath} from cache.")
        return self._cache[abspath]["json"]  # type: ignore

    def _put(self, abspath: str, *, json: str, modified: float):
        self._cache[abspath] = {
            "json": json,
            "last_modified": os.path.getmtime(abspath),
        }
        while self.size > self._max_size:
            key = self._queue.pop()
            self.invalidate_entry(str(key))

    @async_and_sync
    async def get_json(self, file_path: str | Path, *, no_cache: bool = False) -> str:
        """
        Fetches json a str from a file, using the cached version if it is still valid.

        :param file_path : The file path to be fetched.
        :no_cache : If true, file will always be read from the file.
        """
        abspath = self._canonical_file_path(file_path)
        if no_cache:
            return await self._read_json(abspath)

        if self.is_valid(abspath):
            return self._get(abspath)

        self._miss_count = self._miss_count + 1
        logger.debug(f"Reading file {abspath} and adding to cache.")
        json = await self._read_json(abspath)
        self._queue.add(abspath)
        self._put(abspath, json=json, modified=os.path.getmtime(abspath))
        return json

    def invalidate_entry(self, file_path: str | Path) -> None:
        """
        Invalidates the cache for a file path, ensuring it will be re-read on the next read.

        :param file_path : The file path to invalidate cache for.
        """
        abspath = self._canonical_file_path(file_path)
        logger.debug(f"Invalidating cache for file {abspath}.")
        if abspath in self._cache:
            del self._cache[abspath]
            self._queue.remove(abspath)

    def is_valid(self, file_path: str | Path) -> bool:
        """
        Checks whether a cache element is valid.

        :param file_path: The file path to check for.

        :returns : Boolean indicating cache validity.
        """
        abspath = self._canonical_file_path(file_path)

        cache = self._cache[abspath]
        if cache is None:
            return False

        if not os.path.exists(abspath):
            return False

        if os.path.getmtime(abspath) > cache["last_modified"]:
            return False

        return True



import datetime
import glob
import importlib.metadata
import logging
import os
import shutil
from hashlib import md5
from pathlib import Path
from typing import Any, Awaitable, Callable

import filetype
import simplejson  # type: ignore
from async_lru import alru_cache
from packaging.version import Version

from aerovaldb.aerovaldb import AerovalDB
from aerovaldb.const import IMG_FILE_EXTS
from aerovaldb.types import AccessType

from ..exceptions import UnsupportedOperation
from ..lock import FakeLock, FileLock
from ..routes import *
from ..utils import (
    async_and_sync,
    build_uri,
    extract_substitutions,
    json_dumps_wrapper,
    parse_formatted_string,
    parse_uri,
    str_to_bool,
    validate_filename_component,
)
from ..utils.filter import filter_heatmap, filter_regional_stats
from ..utils.string_mapper import StringMapper, VersionConstraintMapper
from .cache import JSONLRUCache

logger = logging.getLogger(__name__)


class AerovalJsonFileDB(AerovalDB):
    def __init__(self, basedir: str | Path, /, use_async: bool = False):
        """
        :param basedir The root directory where aerovaldb will look for files.
        :param asyncio Whether to use asynchronous io to read and store files.
        """
        self._use_real_lock = str_to_bool(
            os.environ.get("AVDB_USE_LOCKING", ""), default=False
        )
        logger.info(
            f"Initializing aerovaldb for '{basedir}' with locking {self._use_real_lock}"
        )

        self._asyncio = use_async
        self._cache = JSONLRUCache(max_size=64, asyncio=self._asyncio)

        self._basedir = os.path.abspath(basedir)

        if not os.path.exists(self._basedir):
            os.makedirs(self._basedir)

        self.PATH_LOOKUP = StringMapper(
            {
                ROUTE_GLOB_STATS: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                ROUTE_REG_STATS: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                ROUTE_HEATMAP: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                # For MAP_OVERLAY, extension is excluded but it will be appended after the fact.
                ROUTE_MAP_OVERLAY: "./{project}/{experiment}/contour/overlay/{source}_{variable}_{date}",
                ROUTE_CONTOUR: "./{project}/{experiment}/contour/{obsvar}_{model}.geojson",
                ROUTE_TIMESERIES_WEEKLY: "./{project}/{experiment}/ts/diurnal/{location}_{network}-{obsvar}_{layer}.json",
                ROUTE_TIMESERIES: "./{project}/{experiment}/ts/{location}_{network}-{obsvar}_{layer}.json",
                ROUTE_EXPERIMENTS: "./{project}/experiments.json",
                ROUTE_CONFIG: "./{project}/{experiment}/cfg_{project}_{experiment}.json",
                ROUTE_MENU: "./{project}/{experiment}/menu.json",
                ROUTE_STATISTICS: "./{project}/{experiment}/statistics.json",
                ROUTE_RANGES: "./{project}/{experiment}/ranges.json",
                ROUTE_REGIONS: "./{project}/{experiment}/regions.json",
                ROUTE_MODELS_STYLE: [
                    "./{project}/{experiment}/models-style.json",
                    "./{project}/models-style.json",
                ],
                ROUTE_MAP: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/map/{network}-{obsvar}_{layer}_{model}-{modvar}_{time}.json",
                        min_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/map/{network}-{obsvar}_{layer}_{model}-{modvar}.json",
                        max_version="0.13.2",
                    ),
                ],
                ROUTE_SCATTER: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/scat/{network}-{obsvar}_{layer}_{model}-{modvar}_{time}.json",
                        min_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/scat/{network}-{obsvar}_{layer}_{model}-{modvar}.json",
                        max_version="0.13.2",
                    ),
                ],
                ROUTE_PROFILES: "./{project}/{experiment}/profiles/{location}_{network}_{obsvar}.json",
                ROUTE_HEATMAP_TIMESERIES: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/{region}-{network}-{obsvar}-{layer}.json",
                        min_version="0.13.2",  # https://github.com/metno/pyaerocom/blob/4478b4eafb96f0ca9fd722be378c9711ae10c1f6/setup.cfg
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/{network}-{obsvar}-{layer}.json",
                        min_version="0.12.2",
                        max_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/stats_ts.json",
                        max_version="0.12.2",
                    ),
                ],
                ROUTE_FORECAST: "./{project}/{experiment}/forecast/{region}_{network}-{obsvar}_{layer}.json",
                ROUTE_GRIDDED_MAP: "./{project}/{experiment}/contour/{obsvar}_{model}.json",
                ROUTE_REPORT: "./reports/{project}/{experiment}/{title}.json",
                ROUTE_REPORT_IMAGE: "./reports/{project}/{experiment}/{path}",
            },
            version_provider=self._get_version,
        )

        self.FILTERS: dict[str, Callable[..., Awaitable[Any]]] = {
            ROUTE_REG_STATS: filter_regional_stats,
            ROUTE_HEATMAP: filter_heatmap,
        }

    @async_and_sync
    @alru_cache(maxsize=2048)
    async def _get_version(self, project: str, experiment: str) -> Version:
        """
        Returns the version of pyaerocom used to generate the files for a given project
        and experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.

        :return : A Version object.
        """
        try:
            config = await self.get_config(project, experiment)
        except FileNotFoundError:
            try:
                # If pyaerocom is installed in the current environment, but no config has
                # been written, we use the version of the installed pyaerocom. This is
                # important for tests to work correctly, and for files to be written
                # correctly if the config file happens to be written after data files.
                version = Version(importlib.metadata.version("pyaerocom"))
            except importlib.metadata.PackageNotFoundError:
                version = Version("0.0.1")
            finally:
                return version

        try:
            version_str = config["exp_info"]["pyaerocom_version"]
            version = Version(version_str)
        except KeyError:
            version = Version("0.0.1")

        return version

    @async_and_sync
    async def _get_template(self, route: str, substitutions: dict) -> str:
        """
        Loops through each instance of TemplateMapper finding the
        appropriate template to use give an route, and a dictionary
        of substitutions.

        :param route : The route for which to look up the template.
        :param substitutions : Dictionary of format substitutions available.

        :returns The template string.

        :raises TemplateNotFound :
            If no valid template was found.
        """
        return await self.PATH_LOOKUP.lookup(route, **substitutions)

    async def _get(
        self,
        route,
        route_args,
        **kwargs,
    ):
        validate_args = kwargs.pop("validate_args", True)
        use_caching = kwargs.pop("cache", False)
        default = kwargs.pop("default", None)
        _raise_file_not_found_error = kwargs.pop("_raise_file_not_found_error", True)
        access_type = self._normalize_access_type(kwargs.pop("access_type", None))

        substitutions = route_args | kwargs
        if validate_args:
            [validate_filename_component(x) for x in substitutions.values()]

        logger.debug(f"Fetching data for {route}.")

        path_template = await self._get_template(route, substitutions)
        logger.debug(f"Using template string {path_template}")

        relative_path = path_template.format(**substitutions)

        file_path = str(Path(os.path.join(self._basedir, relative_path)))
        logger.debug(f"Fetching file {file_path} as {access_type}-")

        filter_func = self.FILTERS.get(route, None)
        filter_vars = route_args | kwargs

        if not os.path.exists(file_path):
            if default is None or access_type == AccessType.FILE_PATH:
                if _raise_file_not_found_error:
                    raise FileNotFoundError(f"File {file_path} does not exist.")
                else:
                    return file_path
            return default

        if access_type in [AccessType.URI]:
            return build_uri(route, route_args, kwargs)

        if filter_func is None:
            if access_type == AccessType.FILE_PATH:
                return file_path

            if access_type == AccessType.JSON_STR:
                raw = await self._cache.get_json(file_path, no_cache=not use_caching)
                return raw

            if access_type == AccessType.MTIME:
                return datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
            if access_type == AccessType.CTIME:
                return datetime.datetime.fromtimestamp(os.path.getctime(file_path))

            raw = await self._cache.get_json(file_path, no_cache=not use_caching)

            return simplejson.loads(raw, allow_nan=True)

        if access_type == AccessType.FILE_PATH:
            raise UnsupportedOperation("Filtered endpoints can not return a filepath")

        json = await self._cache.get_json(file_path, no_cache=not use_caching)
        obj = simplejson.loads(json, allow_nan=True)

        obj = filter_func(obj, **filter_vars)

        if access_type == AccessType.OBJ:
            return obj

        if access_type == AccessType.JSON_STR:
            return json_dumps_wrapper(obj)

        raise UnsupportedOperation

    async def _put(self, obj, route, route_args, **kwargs):
        """Jsondb implemention of database put operation.

        If obj is string, it is assumed to be a wellformatted json string.
        Otherwise it is assumed to be a serializable python object.
        """
        substitutions = route_args | kwargs
        [validate_filename_component(x) for x in substitutions.values()]

        path_template = await self._get_template(route, substitutions)
        relative_path = path_template.format(**substitutions)

        file_path = str(Path(os.path.join(self._basedir, relative_path)).resolve())

        logger.debug(f"Mapped route {route} / { route_args} to file {file_path}.")

        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))
        if isinstance(obj, str):
            json = obj
        else:
            json = json_dumps_wrapper(obj)
        with open(file_path, "w") as f:
            f.write(json)

    def rm_experiment_data(self, project: str, experiment: str) -> None:
        """Deletes ALL data associated with an experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.
        """
        exp_dir = os.path.join(self._basedir, project, experiment)

        if os.path.exists(exp_dir):
            logger.info(
                f"Removing experiment data for project {project}, experiment, {experiment}."
            )
            shutil.rmtree(exp_dir)

    @async_and_sync
    async def get_regional_stats(
        self,
        project: str,
        experiment: str,
        frequency: str,
        network: str,
        variable: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Fetches regional stats from the database.

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param network: Observation network.
        :param variable: Variable name.
        """
        return await self._get(
            ROUTE_REG_STATS,
            {"project": project, "experiment": experiment, "frequency": frequency},
            access_type=kwargs.get("access_type", AccessType.OBJ),
            network=network,
            variable=variable,
            layer=layer,
            cache=True,
        )

    @async_and_sync
    async def get_heatmap(
        self,
        project: str,
        experiment: str,
        frequency: str,
        region: str,
        time: str,
        /,
        *args,
        **kwargs,
    ):
        """Fetches heatmap data from the database

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param region: Region.
        :param time: Time.
        """
        return await self._get(
            ROUTE_HEATMAP,
            {"project": project, "experiment": experiment, "frequency": frequency},
            access_type=kwargs.get("access_type", AccessType.OBJ),
            region=region,
            time=time,
            cache=True,
        )

    @async_and_sync
    async def _get_uri_for_file(self, file_path: str) -> str:
        """
        For the provided data file path, returns the corresponding
        URI.

        :param file_path : The file_path.
        """
        file_path = os.path.join(self._basedir, file_path)
        file_path = os.path.relpath(file_path, start=self._basedir)

        _, ext = os.path.splitext(file_path)

        if "/contour/overlay/" in file_path:
            file_path = str(Path(file_path).parent / Path(file_path).stem)

        if file_path.startswith("reports/") and ext.lower() in IMG_FILE_EXTS:
            # TODO: Fix this.
            # The image endpoint is the only endpoint which needs to accept an arbitrary path
            # under the experiment directory. Treating it as a special case for now.
            split = file_path.split("/")
            project = split[1]
            experiment = split[2]
            path = ":".join(split[3:])
            uri = build_uri(
                ROUTE_REPORT_IMAGE,
                {"project": project, "experiment": experiment, "path": path},
                {},
            )
            return uri

        for route in self.PATH_LOOKUP._lookuptable:
            if not (route == ROUTE_MODELS_STYLE):
                if file_path.startswith("reports/"):
                    _str = "/".join(file_path.split("/")[1:3])
                    subs = parse_formatted_string("{project}/{experiment}", _str)
                else:
                    _str = "/".join(file_path.split("/")[0:2])
                    subs = parse_formatted_string("{project}/{experiment}", _str)
            else:
                try:
                    subs = parse_formatted_string(
                        "{project}/{experiment}/models-style.json", file_path
                    )
                except Exception:
                    try:
                        subs = parse_formatted_string(
                            "{project}/models-style.json", file_path
                        )
                    except:
                        continue

            template = await self._get_template(route, subs)

            if "experiment" in subs:
                version = await self._get_version(subs["project"], subs["experiment"])
            else:
                # Project level models style does not have a version because version is defined
                # per experiment. version doesn't matter for models-style because it is priority
                # based, so we set a dummy value to simplify.
                version = Version("0.0.1")
            route_arg_names = extract_substitutions(route)

            try:
                all_args = parse_formatted_string(template, f"./{file_path}")  # type: ignore
                for k, v in all_args.items():
                    all_args[k] = v.replace("/", ":")

                route_args = {k: v for k, v in all_args.items() if k in route_arg_names}
                kwargs = {
                    k: v for k, v in all_args.items() if not (k in route_arg_names)
                }
            except Exception:
                continue
            else:
                uri = build_uri(route, route_args, kwargs | {"version": version})
                return uri

        raise ValueError(f"Unable to build URI for file path {file_path}")

    @async_and_sync
    async def list_glob_stats(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ):
        access_type = self._normalize_access_type(access_type)
        if access_type in [AccessType.OBJ, AccessType.JSON_STR]:
            raise UnsupportedOperation(f"Unsupported accesstype, {access_type}")

        template = str(
            os.path.abspath(
                os.path.join(
                    self._basedir,
                    await self._get_template(
                        ROUTE_GLOB_STATS, {"project": project, "experiment": experiment}
                    ),  # type: ignore
                )
            )
        )
        glb = template.replace("{frequency}", "*")

        glb = glb.format(project=project, experiment=experiment)

        result = []
        for f in glob.glob(glb):
            if access_type == AccessType.FILE_PATH:
                result.append(f)
                continue

            result.append(await self._get_uri_for_file(f))

        return result

    @async_and_sync
    async def list_timeseries(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ):
        access_type = self._normalize_access_type(access_type)
        if access_type in [AccessType.OBJ, AccessType.JSON_STR]:
            raise UnsupportedOperation(f"Unsupported accesstype, {access_type}")

        template = str(
            os.path.abspath(
                os.path.join(
                    self._basedir,
                    await self._get_template(
                        ROUTE_TIMESERIES,
                        {"project": project, "experiment": experiment},
                    ),  # type: ignore
                )
            )
        )
        glb = (
            template.replace("{location}", "*")
            .replace("{network}", "*")
            .replace("{obsvar}", "*")
            .replace("{layer}", "*")
        )
        glb = glb.format(project=project, experiment=experiment)

        result = []
        for f in glob.glob(glb):
            if access_type == AccessType.FILE_PATH:
                result.append(f)
                continue

            result.append(await self._get_uri_for_file(f))

        return result

    @async_and_sync
    async def list_map(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ):
        access_type = self._normalize_access_type(access_type)
        if access_type in [AccessType.OBJ, AccessType.JSON_STR]:
            raise UnsupportedOperation(f"Unsupported accesstype, {access_type}")

        template = str(
            os.path.abspath(
                os.path.join(
                    self._basedir,
                    self._get_template(
                        ROUTE_MAP,
                        {"project": project, "experiment": experiment},
                    ),  # type: ignore
                )
            )
        )
        glb = (
            template.replace("{network}", "*")
            .replace("{obsvar}", "*")
            .replace("{layer}", "*")
            .replace("{model}", "*")
            .replace("{modvar}", "*")
            .replace("{time}", "*")
        )
        glb = glb.format(project=project, experiment=experiment)

        result = []
        for f in glob.glob(glb):
            if access_type == AccessType.FILE_PATH:
                result.append(f)
                continue

            result.append(await self._get_uri_for_file(f))

        return result

    @async_and_sync
    async def get_by_uri(
        self,
        uri: str,
        /,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
    ):
        access_type = self._normalize_access_type(access_type)
        if access_type in [AccessType.URI]:
            return uri

        route, route_args, kwargs = parse_uri(uri)

        if route.startswith("/v0/report-image/"):
            return await self.get_report_image(
                route_args["project"],
                route_args["experiment"],
                route_args["path"],
                access_type=access_type,
            )

        if route.startswith("/v0/map-overlay/"):
            return await self.get_map_overlay(
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
                access_type=access_type,
            )

        return await self._get(
            route,
            route_args,
            cache=cache,
            default=default,
            access_type=access_type,
            **kwargs,
        )

    @async_and_sync
    async def put_by_uri(self, obj, uri: str):
        route, route_args, kwargs = parse_uri(uri)

        if route.startswith("/v0/report-image/"):
            await self.put_report_image(
                obj, route_args["project"], route_args["experiment"], route_args["path"]
            )
            return

        if route.startswith("/v0/map-overlay/"):
            await self.put_map_overlay(
                obj,
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
            )
            return

        await self._put(obj, route, route_args, **kwargs)

    def _get_lock_file(self) -> str:
        os.makedirs(os.path.expanduser("~/.aerovaldb/.lock/"), exist_ok=True)
        lock_file = os.path.join(
            os.environ.get("AVDB_LOCK_DIR", os.path.expanduser("~/.aerovaldb/.lock/")),
            md5(self._basedir.encode()).hexdigest(),
        )
        return lock_file

    def lock(self):
        if self._use_real_lock:
            return FileLock(self._get_lock_file())

        return FakeLock()

    @async_and_sync
    async def list_all(self, access_type: str | AccessType = AccessType.URI):
        access_type = self._normalize_access_type(access_type)

        if access_type in [AccessType.OBJ, AccessType.JSON_STR]:
            UnsupportedOperation(f"Accesstype {access_type} not supported.")

        glb = glob.iglob(os.path.join(self._basedir, "./**"), recursive=True)

        result = []
        for f in glb:
            if os.path.isfile(f):
                if access_type == AccessType.FILE_PATH:
                    result.append(f)
                    continue

                try:
                    uri = await self._get_uri_for_file(f)
                except (ValueError, KeyError):
                    continue
                else:
                    result.append(uri)

        return result

    @async_and_sync
    async def get_report_image(
        self,
        project: str,
        experiment: str,
        path: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in (
            AccessType.FILE_PATH,
            AccessType.BLOB,
            AccessType.MTIME,
            AccessType.CTIME,
        ):
            raise UnsupportedOperation(
                f"The report image endpoint does not support access type {access_type}."
            )

        if access_type in (AccessType.MTIME, AccessType.CTIME):
            return await self._get(
                route=ROUTE_REPORT_IMAGE,
                route_args={
                    "project": project,
                    "experiment": experiment,
                    "path": path,
                },
                access_type=access_type,
                validate_args=False,
            )
        file_path = await self._get(
            route=ROUTE_REPORT_IMAGE,
            route_args={
                "project": project,
                "experiment": experiment,
                "path": path,
            },
            access_type=AccessType.FILE_PATH,
            validate_args=False,
        )
        logger.debug(f"Fetching image with path '{file_path}'")

        if access_type == AccessType.FILE_PATH:
            return file_path

        with open(file_path, "rb") as f:
            return f.read()

    @async_and_sync
    async def put_report_image(self, obj, project: str, experiment: str, path: str):
        template = await self._get_template(ROUTE_REPORT_IMAGE, {})

        file_path = os.path.join(
            self._basedir,
            template.format(project=project, experiment=experiment, path=path),
        )
        os.makedirs(Path(file_path).parent, exist_ok=True)
        with open(file_path, "wb") as f:
            f.write(obj)

    @async_and_sync
    async def get_map_overlay(
        self,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in (
            AccessType.FILE_PATH,
            AccessType.BLOB,
            AccessType.MTIME,
            AccessType.CTIME,
        ):
            raise UnsupportedOperation(
                f"The report image endpoint does not support access type {access_type}."
            )

        for ext in IMG_FILE_EXTS:
            try:
                file_path = await self._get(
                    route=ROUTE_MAP_OVERLAY,
                    route_args={
                        "project": project,
                        "experiment": experiment,
                        "source": source,
                        "variable": variable,
                        "date": date,
                    },
                    _raise_file_not_found_error=False,
                    access_type=AccessType.FILE_PATH,
                )
            except FileNotFoundError:
                pass
            file_path += ext
            if os.path.exists(file_path):
                break

        logger.debug(f"Fetching image with path '{file_path}'")

        if access_type in [AccessType.MTIME]:
            return datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
        if access_type in [AccessType.CTIME]:
            return datetime.datetime.fromtimestamp(os.path.getctime(file_path))

        if access_type == AccessType.FILE_PATH:
            return file_path

        with open(file_path, "rb") as f:
            return f.read()

    @async_and_sync
    async def put_map_overlay(
        self,
        obj,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
    ):
        """Putter for map overlay images.

        :param obj : The object to be stored.
        :param project : Project ID.
        :param experiment : Experiment ID.
        :param source : Data source. Can be either an observation network or a model ID.
        :param variable : Variable name.
        :param date : Date.
        """
        template = await self._get_template(ROUTE_MAP_OVERLAY, {})

        file_path = os.path.join(
            self._basedir,
            template.format(
                project=project,
                experiment=experiment,
                source=source,
                variable=variable,
                date=date,
            ),
        )

        ext = filetype.guess_extension(obj)
        file_path += f".{ext}"

        Path(file_path).parent.mkdir(exist_ok=True, parents=True)
        with open(file_path, "wb") as f:
            f.write(obj)



import abc
import datetime
import functools
import inspect

from .routes import *
from .types import AccessType
from .utils import async_and_sync


def get_method(route):
    """Decorator for put functions, converts positional-only arguments into route_args

    :param route: the route for this method
    :param wrapped: function template, the wrapper function will never be called
    :return: getter function
    """

    def wrap(wrapped):
        @functools.wraps(wrapped)
        async def wrapper(self, *args, **kwargs):
            sig = inspect.signature(wrapped)
            route_args = {}
            for pos, par in enumerate(sig.parameters):
                if pos == 0:
                    # first time, "self", skip
                    continue
                if sig.parameters[par].kind == inspect.Parameter.POSITIONAL_ONLY:
                    try:
                        route_args[par] = args[0]
                        args = args[1:]
                    except IndexError as iex:
                        raise IndexError(
                            f"{wrapped.__name__} got less parameters as expected (>= {len(route_args)+2}): {iex}"
                        )
            if len(args) > 0:
                raise IndexError(f"{len(args)} superfluous positional args provided.")
            return await self._get(route, route_args, *args, cache=False, **kwargs)

        return wrapper

    return wrap


def put_method(route):
    """Decorator for put functions, converts positional-only arguments into route_args

    :param route: the route for this method
    :param wrapped: function template, the function will never be called
    :return: putter function
    """

    def wrap(wrapped):
        @functools.wraps(wrapped)
        async def wrapper(self, obj, *args, **kwargs):
            sig = inspect.signature(wrapped)
            route_args = {}
            for pos, par in enumerate(sig.parameters):
                if pos <= 1:
                    # first time, "self", "obj" skip
                    continue
                if sig.parameters[par].kind == inspect.Parameter.POSITIONAL_ONLY:
                    try:
                        route_args[par] = args[0]
                        args = args[1:]
                    except IndexError as iex:
                        raise IndexError(
                            f"{wrapped.__name__} got less parameters as expected (>= {len(route_args)+2}): {iex}"
                        )
            if len(args) > 0:
                raise IndexError(f"{len(args)} superfluous positional args provided.")
            return await self._put(obj, route, route_args, **kwargs)

        return wrapper

    return wrap


class AerovalDB(abc.ABC):
    def __init__(self):
        """AerovalDB is the base class for databases. Please make sure to initialize
        any instance of aerovaldb as a context manager to ensure resources to be release,
        e.g. database-handles.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def _get(self, route: str, route_args: dict[str, str], **kwargs):
        """Abstract implementation of the main getter functions. All get and put
        functions map to this function, with a corresponding route as key
        to enable key/value pair put and get functionality.

        :param route: a route similar to a REST route
        :param route_args: route parameters, like format-dicts
        """
        raise NotImplementedError

    async def _put(self, obj, route: str, route_args: dict[str, str], **kwargs):
        """Abstract implementation of the main getter functions. All get and put
        functions map to this function, with a corresponding route as key
        to enable key/value pair put and get functionality.

        :param obj: the object to put
        :param route: a route similar to a REST route
        :param route_args: route parameters, like format-dicts
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_GLOB_STATS)
    async def get_glob_stats(
        self,
        project: str,
        experiment: str,
        frequency: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a glob_stats object from the database.

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency (eg. 'monthly')
        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_REG_STATS)
    async def get_regional_stats(
        self,
        project: str,
        experiment: str,
        frequency: str,
        network: str,
        variable: str,
        layer: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches regional stats from the database.

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param network: Observation network.
        :param variable: Variable name.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_HEATMAP)
    async def get_heatmap(
        self,
        project: str,
        experiment: str,
        frequency: str,
        region: str,
        time: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches heatmap data from the database

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param region: Region.
        :param time: Time.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_GLOB_STATS)
    async def put_glob_stats(
        self, obj, project: str, experiment: str, frequency: str, /, *args, **kwargs
    ):
        """Saves a glob_stats object to the database.

        :param obj: The object to be stored.
        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency (eg. 'monthly')
        """
        raise NotImplementedError

    @async_and_sync
    async def list_glob_stats(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ) -> list[str]:
        """Lists the URI for each glob_stats object.

        :param project: str
        :param experiment: str

        :return List of URIs.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_CONTOUR)
    async def get_contour(
        self,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetch a contour object from the db.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param obsvar: Observation variable.
        :param model: Model ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_CONTOUR)
    async def put_contour(
        self,
        obj,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        **kwargs,
    ):
        """Put a contour object in the db.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param obsvar: Observation variable.
        :param model: Model ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_TIMESERIES)
    async def get_timeseries(
        self,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a timeseries from the db.

        :param project: Project id.
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_TIMESERIES)
    async def put_timeseries(
        self,
        obj,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Places a timeseries in the db

        :param obj: The object to write into the db.
        :param project: Project ID
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        """
        raise NotImplementedError

    @async_and_sync
    async def list_timeseries(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ) -> list[str]:
        """Returns a list of URIs of all timeseries files for
        a given project and experiment id.

        :param project : Project ID.
        :param experiment : Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_TIMESERIES_WEEKLY)
    async def get_timeseries_weekly(
        self,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a weekly time series from the db.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_TIMESERIES_WEEKLY)
    async def put_timeseries_weekly(
        self,
        obj,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Stores a weekly time series in the db.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_EXPERIMENTS)
    async def get_experiments(
        self,
        project: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a list of experiments for a project from the db.

        :param project: Project ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_EXPERIMENTS)
    async def put_experiments(self, obj, project: str, /, *args, **kwargs):
        """Stores a list of experiments for a project from the db.

        :param project: Project ID.
        """
        raise NotImplementedError

    def rm_experiment_data(self, project: str, experiment: str):
        """Deletes ALL data associated with an experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.
        """
        raise NotImplementedError

    def _list_experiments(
        self, project: str, /, has_results: bool = False
    ) -> list[str]:
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_CONFIG)
    async def get_config(
        self,
        project: str,
        experiment: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a configuration from the db.

        :param project: Project ID.
        :param experiment: Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_CONFIG)
    async def put_config(self, obj, project: str, experiment: str, /, *args, **kwargs):
        """Stores a configuration to the db.

        :paran obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        """

        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_MENU)
    async def get_menu(
        self,
        project: str,
        experiment: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches a menu configuartion from the db.

        :param project: Project ID.
        :param experiment: Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_MENU)
    async def put_menu(self, obj, project: str, experiment: str, /, *args, **kwargs):
        """Stores a menu configuration in the db.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_STATISTICS)
    async def get_statistics(
        self,
        project: str,
        experiment: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches statistics for an experiment.

        :param project: Project ID.
        :param experiment: Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_STATISTICS)
    async def put_statistics(
        self, obj, project: str, experiment: str, /, *args, **kwargs
    ):
        """Stores statistics to the db.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_RANGES)
    async def get_ranges(
        self,
        project: str,
        experiment: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches ranges from the db.

        :param project: Project ID.
        :param experiment: Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_RANGES)
    async def put_ranges(self, obj, project: str, experiment: str, /, *args, **kwargs):
        """Stores ranges in db.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_REGIONS)
    async def get_regions(
        self,
        project: str,
        experiment: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches regions from db.

        :param project: Project ID.
        :param experiment: Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_REGIONS)
    async def put_regions(self, obj, project: str, experiment: str, /, *args, **kwargs):
        """Stores regions in db.

        :param obj: Object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_MODELS_STYLE)
    async def get_models_style(
        self,
        project: str,
        /,
        experiment: str | None = None,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches model styles from db.

        :param project: Project ID.
        :param experiment (Optional): Experiment ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_MODELS_STYLE)
    async def put_models_style(
        self, obj, project: str, /, experiment: str | None = None, *args, **kwargs
    ):
        """Stores model styles config in db.

        :param obj: Object to be stored.
        :param project: Project ID.
        :param experiment (Optional): Experiment ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_MAP)
    async def get_map(
        self,
        project: str,
        experiment: str,
        network: str,
        obsvar: str,
        layer: str,
        model: str,
        modvar: str,
        time: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches map data from db.

        :param project: Project Id
        :param experiment: Experiment ID.
        :param network: Observation network
        :param obsvar: Observation variable.
        :param layer: Layer
        :param model: Model ID
        :param modvar: Model variable.
        :param time: Time parameter.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_MAP)
    async def put_map(
        self,
        obj,
        project: str,
        experiment: str,
        network: str,
        obsvar: str,
        layer: str,
        model: str,
        modvar: str,
        time: str,
        /,
        *args,
        **kwargs,
    ):
        """Stores map data in db.

        :param obj: The Object to be stored.
        :param project: Project Id
        :param experiment: Experiment ID.
        :param network: Observation network
        :param obsvar: Observation variable.
        :param layer: Layer
        :param model: Model ID
        :param modvar: Model variable.
        :param time: Time parameter.
        """
        raise NotImplementedError

    @async_and_sync
    async def list_map(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ) -> list[str]:
        """Lists all map files for a given project / experiment combination.

        :param project: The project ID.
        :param experiment: The experiment ID.

        :return List with the URIs.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_SCATTER)
    async def get_scatter(
        self,
        project: str,
        experiment: str,
        network: str,
        obsvar: str,
        layer: str,
        model: str,
        modvar: str,
        time: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Get scat.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        :param model: Model ID.
        :param modvar: Model variable.
        :param time: Time parameter.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_SCATTER)
    async def put_scatter(
        self,
        obj,
        project: str,
        experiment: str,
        network: str,
        obsvar: str,
        layer: str,
        model: str,
        modvar: str,
        time: str,
        /,
        *args,
        **kwargs,
    ):
        """Stores scat in db.

        :param obj: Object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param network: Observation network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        :param model: Model ID.
        :param modvar: Model variable.
        :param time: Time paramter.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_PROFILES)
    async def get_profiles(
        self,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches profiles from db.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_PROFILES)
    async def put_profiles(
        self,
        obj,
        project: str,
        experiment: str,
        location: str,
        network: str,
        obsvar: str,
        /,
        *args,
        **kwargs,
    ):
        """Stores profiles in db.

        :param obj: Object to be stored.
        :param project: Project ID._
        :param experiment: Experiment ID.
        :param location: Location.
        :param network: Observation network.
        :param obsvar: Observation variable.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_HEATMAP_TIMESERIES)
    async def get_heatmap_timeseries(
        self,
        project: str,
        experiment: str,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches heatmap timeseries.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param region: Region ID.
        :param network: Observation Network.
        :param obsvar: Observation variable.
        :param layer: Layer.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_HEATMAP_TIMESERIES)
    async def put_heatmap_timeseries(
        self,
        obj,
        project: str,
        experiment: str,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Stores heatmap timeseries.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param region: Region ID.
        :param network: Observation Network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_FORECAST)
    async def get_forecast(
        self,
        project: str,
        experiment: str,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetch forecast.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param region: Region ID.
        :param network: Observation Network.
        :param obsvar: Observation variable.
        :param layer: Layer.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_FORECAST)
    async def put_forecast(
        self,
        obj,
        project: str,
        experiment: str,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Store forecast.

        :param obj: The Object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param region: Region ID.
        :param network: Observation Network.
        :param obsvar: Observation variable.
        :param layer: Layer.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_GRIDDED_MAP)
    async def get_gridded_map(
        self,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetches gridded map.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param obsvar: Observation variable.
        :param model: Model ID.

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_GRIDDED_MAP)
    async def put_gridded_map(
        self,
        obj,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        **kwargs,
    ):
        """Store gridded map.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param obsvar: Observation variable.
        :param model: Model ID.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_REPORT)
    async def get_report(
        self,
        project: str,
        experiment: str,
        title: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        """Fetch report.

        :param project: Project ID.
        :param experiment: Experiment ID.
        :param title: Report title (ie. filename without extension).

        :param access_type: How the data is to be retrieved (See AccessType for details)
        :param cache: Whether to use cache for this read.
        :param default: Default value that will be returned instead of raising FileNotFoundError
        if not data was found (Will be returned as is and not converted to match access_type).

        :returns The fetched data.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_REPORT)
    async def put_report(
        self, obj, project: str, experiment: str, title: str, /, *args, **kwargs
    ):
        """Store report.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param title: Report title (ie. filename without extension).
        """
        raise NotImplementedError

    @async_and_sync
    async def get_by_uri(
        self,
        uri: str,
        /,
        access_type: str | AccessType,
        cache: bool = False,
        default=None,
    ):
        """Gets a stored object by its URI.

        :param uri : URI of the item to fetch.
        :param access_type : See AccessType.
        :param cache : Whether to use the cache.
        :param default : If provided, this value will be returned instead of raising
        a FileNotFoundError if not file exists. The provided object will be returned
        as is, and will not be converted to match access_type.

        Note:
        -----
        URI is intended to be consistent between implementations. Using get_by_uri()
        to fetch an identifier which can then be written to another connector using
        its respective put_by_uri() method.
        """
        raise NotImplementedError

    @async_and_sync
    async def put_by_uri(self, obj, uri: str):
        """Replaces a stored object by uri with a new object.

        :param obj: The object to be stored. Either a json str, or a
        json serializable python object.
        :param uri: The uri as which to store the object.

        Note:
        -----
        URI is intended to be consistent between implementations. Using get_by_uri()
        to fetch an identifier which can then be written to another connector using
        its respective put_by_uri() method.
        """
        raise NotImplementedError

    def lock(self):
        """Acquires an exclusive advisory lock to coordinate file access
        between instances of aerovaldb. Intended to be used as a context
        manager.

        See also: https://aerovaldb.readthedocs.io/en/latest/locking.html
        """
        raise NotImplementedError

    def _normalize_access_type(
        self, access_type: AccessType | str | None, default: AccessType = AccessType.OBJ
    ) -> AccessType:
        """Normalizes the access_type to an instance of AccessType enum.

        :param access_type: AccessType instance or string convertible to AccessType
        :param default: The type to return if access_type is None. Defaults to AccessType.OBJ
        :raises ValueError: If str access_type can't be converted to AccessType.
        :raises ValueError: If access_type is not str or AccessType
        :return: The normalized AccessType.
        """
        if isinstance(access_type, AccessType):
            return access_type

        if isinstance(access_type, str):
            try:
                return AccessType[access_type]
            except:
                raise ValueError(
                    f"String '{access_type}' can not be converted to AccessType."
                )
        if access_type is None:
            return default

        assert False

    @async_and_sync
    async def list_all(self, access_type: str | AccessType = AccessType.URI):
        """Iterator to list over the URI of each object
        stored in the current aerovaldb connection, returning
        the URI of each.

        :param access_type : What to return (This is implementation specific, but in general
        each implementation should support URI).
        :raises : UnsupportedOperation
            For non-supported acces types.
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_REPORT_IMAGE)
    async def get_report_image(
        self,
        project: str,
        experiment: str,
        path: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        """
        Getter for static images that are referenced from the report json files.

        :param project : Project ID.
        :param experiment : Experiment ID.
        :param access_type : One of AccessType.BLOB, AccessType.FILE_PATH

        :return Either a string (If file path requested) or a bytes object with the
            image data
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_REPORT_IMAGE)
    async def put_report_image(self, obj, project: str, experiment: str, path: str):
        """
        Putter for static images that are referenced from the report json files.

        :param obj : A bytes object representing the image data to be written.
        :param project : Project ID.
        :param experiment : Experiment ID.

        :return Either a string (If file path requested) or a bytes object with the
            image data
        """
        raise NotImplementedError

    @async_and_sync
    @get_method(ROUTE_MAP_OVERLAY)
    async def get_map_overlay(
        self,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        """Getter for map overlay images.

        :param project : Project ID.
        :param experiment : Experiment ID.
        :param source : Data source. Can be either an observation network or a model ID.
        :param variable : Variable name.
        :param date : Date.
        """
        raise NotImplementedError

    @async_and_sync
    @put_method(ROUTE_MAP_OVERLAY)
    async def put_map_overlay(
        self,
        obj,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
    ):
        """Putter for map overlay images.

        :param obj : The object to be stored.
        :param project : Project ID.
        :param experiment : Experiment ID.
        :param source : Data source. Can be either an observation network or a model ID.
        :param variable : Variable name.
        :param date : Date.
        """
        raise NotImplementedError

    @async_and_sync
    async def get_experiment_mtime(
        self, project: str, experiment: str
    ) -> datetime.datetime:
        """
        :param project: Project ID.
        :param experiment: Experiment ID.
        """
        uri = await self.get_config(project, experiment, access_type=AccessType.URI)
        return await self.get_by_uri(uri, access_type=AccessType.MTIME)
